@@ -27,6 +27,7 @@ from PIL import Image
 # Self imports
 from .thermogram import FlyrThermogram
 from .palette_info import Palette
+from .picture_in_picture import PictureInPicture
 
 
 # Constants
@@ -203,6 +204,7 @@ def parse_thermal(
     RECORD_IDX_EMBEDDED_IMAGE = 14
     RECORD_IDX_CAMERA_INFO = 32
     RECORD_IDX_PALETTE_INFO = 34
+    RECORD_IDX_PICTURE_IN_PICTURE_INFO = 42
 
     raw_data_md = records[RECORD_IDX_RAW_DATA]
     width, height, raw_data = parse_raw_data(stream, raw_data_md)
@@ -223,8 +225,20 @@ def parse_thermal(
     except:
         palette_info = None
 
+    try:
+        pip_info_md = records[RECORD_IDX_PICTURE_IN_PICTURE_INFO]
+        pip_info: Optional[PictureInPicture] = parse_picture_in_picture_info(
+            stream, pip_info_md
+        )
+    except:
+        pip_info = None
+
     thermogram = FlyrThermogram(
-        raw_data, camera_info, optical=embedded_image, palette=palette_info
+        raw_data,
+        camera_info,
+        optical=embedded_image,
+        palette=palette_info,
+        picture_in_picture=pip_info,
     )
     return thermogram
 
@@ -452,16 +466,12 @@ def parse_camera_info(
         "ir_window_temperature",
         "reflected_apparent_temperature",
     ]
-    camera_info = {
-        k: round(v, 2) if k in to_round else v for k, v in camera_info.items()
-    }
+    camera_info = {k: round(v, 2) if k in to_round else v for k, v in camera_info.items()}
 
     return camera_info
 
 
-def parse_palette_info(
-    stream: BinaryIO, metadata: Tuple[int, int, int, int]
-) -> Palette:
+def parse_palette_info(stream: BinaryIO, metadata: Tuple[int, int, int, int]) -> Palette:
     (_, _, offset, _) = metadata
 
     stream.seek(offset)
@@ -505,6 +515,27 @@ def parse_palette_info(
     )
 
 
+def parse_picture_in_picture_info(
+    stream: BinaryIO, metadata: Tuple[int, int, int, int]
+) -> PictureInPicture:
+    (_, _, offset, _) = metadata
+
+    stream.seek(offset)
+
+    ratio = stream.read(4)
+    real_to_ir = struct.unpack("f", ratio)[0]
+    offset_x = int.from_bytes(stream.read(2), "little", signed=True)
+    offset_y = int.from_bytes(stream.read(2), "little", signed=True)
+    pip_x1 = int.from_bytes(stream.read(2), "little", signed=True)
+    pip_x2 = int.from_bytes(stream.read(2), "little", signed=True)
+    pip_y1 = int.from_bytes(stream.read(2), "little", signed=True)
+    pip_y2 = int.from_bytes(stream.read(2), "little", signed=True)
+
+    return PictureInPicture(
+        real_to_ir, offset_x, offset_y, pip_x1, pip_x2, pip_y1, pip_y2
+    )
+
+
 def main():
     import argparse
     import warnings
@@ -515,8 +546,11 @@ def main():
     parser.add_argument(
         "--camera-info", "-c", action="store_true", default=False, required=False
     )
-    parser.add_argument("--palette-dest", "-p", default="", required=False)
+    parser.add_argument("--palette-dest", default="", required=False)
     parser.add_argument("--render-dest", "-r", default="", required=False)
+    parser.add_argument("--optical-dest", "-o", default="", required=False)
+    parser.add_argument("--pnp-dest", "-p", default="", required=False)
+    parser.add_argument("--emboss-dest", "-e", default="", required=False)
     parser.add_argument("thermogram")
     args = parser.parse_args()
 
@@ -538,6 +572,15 @@ def main():
 
         if args.render_dest:
             thermogram.render_pil().save(args.render_dest)
+
+        if args.optical_dest:
+            thermogram.optical_pil.save(args.optical_dest)
+
+        if args.pnp_dest:
+            thermogram.picture_in_picture().save(args.pnp_dest)
+
+        if args.emboss_dest:
+            thermogram.render_pil(edge_emphasis=0.275).save(args.emboss_dest)
     except Exception as e:
         msg = f"{tb.format_exc()}\n{script}: error: failed unpacking {args.thermogram}"
         print(msg)
