@@ -5,14 +5,16 @@ import os
 import warnings
 import numpy as np
 
-from PIL import Image
+from PIL import Image, ImageFilter
 from nptyping import Array
 from math import sqrt, exp, fsum
-from typing import List, Optional, Dict, Tuple, Union
+from typing import List, Optional, Dict, Tuple, Union, Any
 
 import flyr.palettes as palettes
 import flyr.normalization as norm
-import flyr.palette_info as pinfo
+import flyr.palette_info as pal
+import flyr.picture_in_picture as pip
+import flyr.image_manipulation as manip
 
 
 class FlyrThermogram:
@@ -36,7 +38,8 @@ class FlyrThermogram:
     __metadata: Dict[str, Union[float, int]]
     __metadata_adjustments: Dict[str, Union[float, int]]
     __optical: Optional[Array[int, ..., ..., 3]]
-    palette: Optional[pinfo.Palette]
+    palette: Optional[pal.Palette]
+    pip_info: Optional[pip.PictureInPicture]
     path: Optional[str]
 
     def __init__(
@@ -44,7 +47,8 @@ class FlyrThermogram:
         thermal: Array[np.int64, ..., ...],
         metadata: Dict[str, Union[float, int]],
         optical: Optional[Array[np.uint8, ..., ..., 3]] = None,
-        palette: Optional[pinfo.Palette] = None,
+        palette: Optional[pal.Palette] = None,
+        picture_in_picture: Optional[pip.PictureInPicture] = None,
         path: Optional[str] = None,
         metadata_adjustments: Dict[str, Union[float, int]] = {},
     ):
@@ -68,6 +72,9 @@ class FlyrThermogram:
                 The palette embedded in the file, which can be used to render
                 with by default. Is `None` by default in which case a grayscale
                 palette is used.
+            picture_in_picture: Optional[pip.PictureInPicture]
+                The picture-in-picture settings useful for stacking optical and
+                thermal data.
             path: Optional[str]
                 The path to the original file. Is only used to determine this
                 this object's identifier and can safely be left `None`
@@ -86,6 +93,7 @@ class FlyrThermogram:
         self.__metadata = metadata.copy()
         self.__metadata_adjustments = metadata_adjustments.copy()
         self.palette = palette
+        self.pip_info = picture_in_picture
         self.path = path
 
     @property
@@ -137,7 +145,7 @@ class FlyrThermogram:
         return None if self.__optical is None else self.__optical.copy()
 
     @property
-    def optical_pil(self) -> Image:
+    def optical_pil(self) -> Image.Image:
         """ Returns the thermogram's embedded photo as a Pillow `Image`.
 
             Returns
@@ -234,26 +242,196 @@ class FlyrThermogram:
 
         # Render
         normalized = normalizer[mode](min_v, max_v, self.kelvin)
-        if palette == "grayscale":
+        if palette in ["grayscale", "grayscale-inverted"]:
             # Strategy for grayscale is very different from when using a map
             rendered = (normalized * 255).astype(np.uint8)
             outshape = rendered.shape + (3,)
             repeated = np.broadcast_to(rendered[..., None], outshape)
-            return np.clip(repeated, 0, 255)  # return grayscale
+            repeated = np.clip(repeated, 0, 255)
+            if palette == "grayscale-inverted":
+                repeated = np.invert(repeated)
+            return repeated  # return grayscale
         elif palette == "embedded" and self.palette is not None:
             palette = self.palette.rgbs
         return palettes.map_colors(normalized, palette)  # return with color map
 
-    def render_pil(self, **kwargs) -> Image:
-        """ Renders the thermogram, but returns a pillow Image object. See
-            `render()` for documentation on the parameters and other details.
+    def render_pil(
+        self,
+        edge_emphasis: float = 0.0,
+        mask: Optional[Array[bool, ..., ...]] = None,
+        **kwargs,
+    ) -> Image.Image:
+        """ Renders the thermogram, but returns a pillow Image object.
+
+            See `render()` for documentation on the parameters and other details.
+            The parameters listed below are exclusive for `render_pil()`.
+
+            Parameters
+            ----------
+            edge_emphasis: float
+                Default is `0.0`. The strength with which to enhance the render's
+                edges, using the optical image as a base. The value must be between
+                `0.0` and `1.0`, where the first means no emphasis. No edge emphasis
+                happens in case there is no optical image present.
+            mask: Optional[Array[bool, ..., ...]]
+                Default is `None`. The region where the choosen palette will be applied
+                and the edge emphasis enhancer. The mask must be a boolean type where
+                True refers to the region of interest. Its shape must be the same as
+                the thermal data's.
 
             Returns
             -------
             PIL.Image
                 A pillow Image of the rendered thermogram.
         """
-        return Image.fromarray(self.render(**kwargs))
+        assert mask is None or (
+            isinstance(mask, np.ndarray)
+            and mask.dtype == "bool"
+            and mask.shape == self.kelvin.shape
+        )
+
+        bg_render_pil = fg_render_pil = Image.fromarray(self.render(**kwargs))
+        if mask is not None and np.all(mask) is not True:
+            kwargs["palette"] = "grayscale"
+            bg_render_pil = Image.fromarray(self.render(**kwargs))
+            roi_mask = Image.fromarray(mask).convert("L")
+            roi_mask = roi_mask.filter(ImageFilter.BoxBlur(2))
+            bg_render_pil.paste(fg_render_pil, (0, 0), roi_mask)
+
+        if self.__optical is not None and self.pip_info and edge_emphasis > 0.0:
+            ratio = self.pip_info.real_to_ir
+            offset_x = self.pip_info.offset_x
+            offset_y = self.pip_info.offset_y
+
+            ratio = self.__optical.shape[1] / bg_render_pil.size[0] / ratio
+            src_size = bg_render_pil.size
+            dst_size = (
+                round(bg_render_pil.size[0] * ratio),
+                round(bg_render_pil.size[1] * ratio),
+            )
+
+            height, width = self.__optical.shape[:2]
+            origin_x, origin_y = width // 2, height // 2
+            origin_x = origin_x - dst_size[0] // 2 + offset_x
+            origin_y = origin_y - dst_size[1] // 2 + offset_y
+
+            thermal_surface = (
+                origin_x,
+                origin_y,
+                origin_x + dst_size[0],
+                origin_y + dst_size[1],
+            )
+            embossment, opacity = self.embossment_masks_pil(opacity=edge_emphasis)
+            embossment = embossment.crop(thermal_surface).resize(src_size)
+            opacity = opacity.crop(thermal_surface).resize(src_size)
+
+            if mask is not None and np.all(mask) is not True:
+                embossment = manip.apply_mask_pil(embossment, mask)
+                opacity = manip.apply_mask_pil(opacity, mask)
+
+            bg_render_pil = bg_render_pil.convert("RGBA")
+            bg_render_pil.paste(embossment, (0, 0), opacity)
+        return bg_render_pil.convert("RGB")
+
+    def picture_in_picture_pil(
+        self,
+        render: Optional[Image.Image] = None,
+        render_opacity: float = 1.0,
+        render_crop: bool = True,
+        edge_emphasis: float = 0.0,
+        mask: Optional[Array[bool, ..., ...]] = None,
+        mask_mode: str = "classical",
+    ) -> Image.Image:
+        if self.__optical is None or self.pip_info is None:
+            raise ValueError("Impossible operation due missing information")
+
+        assert mask_mode in ["classical", "alternative"]
+
+        ratio = self.pip_info.real_to_ir
+        offset_x = self.pip_info.offset_x
+        offset_y = self.pip_info.offset_y
+        render_opacity = min(max(render_opacity, 0.0), 1.0)
+
+        # Get the optical and render
+        optical_pil = self.optical_pil.convert("RGBA")
+        render_pil: Image.Image
+        if render is None:
+            # Render thermogram if none given, with mask if alternative mask mode
+            mask_ = mask if mask_mode == "alternative" else None
+            render_pil = self.render_pil(edge_emphasis=edge_emphasis, mask=mask_)
+
+            # Ensure there is an alpha channel and make transparent if classical mask mode
+            render_arr = np.array(render_pil.convert("RGBA"))
+            render_arr[:, :, 3] = round(255 * render_opacity)
+            if mask_mode == "classical" and mask is not None and np.any(np.invert(mask)):
+                render_arr[np.invert(mask)] = 0
+
+            render_pil = Image.fromarray(render_arr)
+        else:
+            render_pil = render.convert("RGBA")
+
+        # Resize the images according the picture-in-picture metadata
+        ratio = optical_pil.size[0] / render_pil.size[0] / ratio
+        if render_crop:
+            render_pil = render_pil.crop(self.pip_info.crop_box)
+
+        dst = (round(render_pil.size[0] * ratio), round(render_pil.size[1] * ratio))
+        render_pil = render_pil.resize(dst)
+
+        # Calculate placement and finish
+        width, height = optical_pil.size
+        origin_x, origin_y = width // 2, height // 2
+        origin_x = origin_x - render_pil.size[0] // 2 + offset_x
+        origin_y = origin_y - render_pil.size[1] // 2 + offset_y
+
+        optical_pil.paste(render_pil, (origin_x, origin_y), render_pil)
+        return optical_pil.convert("RGB")
+
+    def embossment_masks(self, opacity: float = 0.275) -> Tuple[np.array, np.array]:
+        opacity = min(max(opacity, 0.0), 1.0)
+        if self.__optical is None:
+            raise ValueError("Impossible operation due to missing optical image")
+
+        # Define azimuth, elevation, and depth
+        ele = np.pi / 2.2  # radians
+        azi = np.pi / 4.0  # radians
+        dep = 15.0  # (0-100)
+
+        # Get a B&W version of the image
+        img = Image.fromarray(self.__optical).convert("L")
+        arr_embossment = np.array(img).astype(float)
+
+        # Find the gradient and adjust the gradient by the depth factor
+        grad_x, grad_y = np.gradient(arr_embossment)
+        grad_x = grad_x * dep / 100.0
+        grad_y = grad_y * dep / 100.0
+
+        # Get the unit incident ray
+        gd = np.cos(ele)  # length of projection of ray on ground plane
+        dx = gd * np.cos(azi)
+        dy = gd * np.sin(azi)
+        dz = np.sin(ele)
+
+        # Find the unit normal vectors for the image
+        leng = np.sqrt(grad_x ** 2 + grad_y ** 2 + 1.0)
+        uni_x = grad_x / leng
+        uni_y = grad_y / leng
+        uni_z = 1.0 / leng
+
+        # Take the dot product, avoiding overflow
+        arr_embossment = 255 * (dx * uni_x + dy * uni_y + dz * uni_z)
+        arr_embossment = arr_embossment.clip(0, 255).astype(np.uint8)
+
+        # Threshold non-edges to be white and see-through
+        arr_embossment[arr_embossment > 180] = 255
+        arr_embossment = arr_embossment.astype(np.uint8)
+        arr_opacity = np.full_like(arr_embossment, fill_value=round(opacity * 255))
+        arr_opacity[arr_embossment > 180] = 0
+        return arr_embossment, arr_opacity
+
+    def embossment_masks_pil(self, **kwargs) -> Tuple[Image.Image, Image.Image]:
+        e, o = self.embossment_masks(**kwargs)
+        return Image.fromarray(e), Image.fromarray(o)
 
     def adjust_metadata(
         self, in_place=False, **kwargs: Union[float, int]
@@ -319,6 +497,7 @@ class FlyrThermogram:
             self.__metadata.copy(),
             None if self.__optical is None else self.__optical.copy(),
             palette=self.palette,
+            picture_in_picture=self.pip_info,
             metadata_adjustments=metadata_adjustments,
             path=self.path,
         )
@@ -467,13 +646,7 @@ class FlyrThermogram:
         raw_atm2_attn = (1 - tau2) / term3 / tau2 * raw_atm2
 
         subtraction = fsum(
-            [
-                raw_atm1_attn,
-                raw_atm2_attn,
-                raw_wind_attn,
-                raw_refl1_attn,
-                raw_refl2_attn,
-            ]
+            [raw_atm1_attn, raw_atm2_attn, raw_wind_attn, raw_refl1_attn, raw_refl2_attn]
         )
 
         raw_obj = thermal.astype(np.float64)
